#ifndef RFI_TYPES_H
#define RFI_TYPES_H

class AntennaFlagCountPlot;
class FrequencyFlagCountPlot;
class FrequencyPowerPlot;
class PolarizationStatistics;
class TimeFlagCountPlot;
class IterationsPlot;

#endif // RFI_TYPES_H
